# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Create default notes for Scade models."""

from enum import Enum
from typing import Dict

import scade.model.suite as suite
from scade.model.suite import get_roots as get_sessions
import scade.model.suite.annotation as ann
from scade.model.suite.visitors import Visit

tool = 'Ansys SCADE Power Scripts: Create default notes for Scade models'


class FT(Enum):
    """Annotation attribute type codes."""

    # values obtained by reverse engineering
    INTEGER = 1
    REAL = 2
    BOOLEAN = 3
    CHARACTER = 4
    STRING = 5
    TEXT = 6
    DATE = 8
    ENUM = 9
    FILE = 10


class DefaultNotes(Visit):
    """Creation of default notes for a model."""

    def __init__(self):
        # runtime properties
        self.categories: Dict[str, ann.AnnCategory] = {}
        self.defaults: Dict[ann.AnnAttDefinition, str] = {}

    def build_categories(self, model: suite.Model):
        """Create a map of all categories allowing a faster access."""
        # retrieve the global annotation schema:
        # since it is not directly accessible from the model
        # we need to retrieve it from any note type
        for note_type in model.ann_note_types:
            schema = note_type.annot_schema
            break
        else:
            # no annotation schema: there is nothing to do
            return

        self.categories = {_.name: _ for _ in schema.ann_categories}

    def get_category(self, annotable: suite.Annotable) -> str:
        """Retrieve the category of an annotable element."""
        # default
        category = ''

        if isinstance(annotable, suite.Package) and not isinstance(annotable, suite.Model):
            category = 'package'
        elif isinstance(annotable, suite.Constant):
            category = 'type_element' if annotable.enumeration else 'constant'
        elif isinstance(annotable, suite.LocalVariable):
            if annotable.is_input():
                category = 'input'
            elif annotable.is_output():
                category = 'output'
            elif annotable.is_hidden():
                category = 'hidden'
            elif annotable.is_local() and not annotable.is_internal():
                category = 'probe' if annotable.probe else 'local'
            elif annotable.is_signal():
                category = 'signal'
        elif isinstance(annotable, suite.Sensor):
            category = 'sensor'
        elif isinstance(annotable, suite.NamedType):
            category = 'named_type'
        elif isinstance(annotable, suite.CompositeElement):
            category = 'type_element'
        elif isinstance(annotable, suite.Operator):
            category = 'operator'
        elif isinstance(annotable, suite.Assertion):
            category = 'assertion'
        elif isinstance(annotable, suite.State):
            category = 'state'
        elif isinstance(annotable, suite.StateMachine):
            category = 'state_machine'
        elif isinstance(annotable, suite.Transition):
            category = 'transition'
        elif isinstance(annotable, suite.ActivateBlock):
            category = 'activate_block'
        elif isinstance(annotable, suite.IfNode):
            category = 'branch'
        # when branches can't be annotated in the iDE
        # elif isinstance(annotable, suite.WhenBranch):
        #     category = 'branch'
        elif isinstance(annotable, suite.Action):
            category = 'action'
        elif isinstance(annotable, suite.Equation):
            if annotable.terminator:
                category = 'terminator'
            else:
                ge = annotable.equation_ge
                if ge and ge.kind == 'OBJ_LIT':
                    category = 'literal'
                else:
                    right = annotable.right
                    if isinstance(right, suite.ExprCall):
                        category = 'ref_operator' if right.operator else 'predef_opr'
                    elif isinstance(right, suite.ExprId):
                        category = 'ref_ident'
        elif isinstance(annotable, suite.EquationSet):
            category = 'equation_set'
        elif isinstance(annotable, suite.NetDiagram):
            category = 'net_diagram'
        elif isinstance(annotable, suite.TextDiagram):
            category = 'text_diagram'
        # documented but meaningless: can't be edited
        # elif isinstance(annotable, suite.TreeDiagram):
        #     category = 'tree_diagram'

        return category

    def compute_default_values(self, model: suite.Model):
        """Update the map of default values for note attributes."""
        for note_type in model.ann_note_types:
            for att in note_type.ann_att_definitions:
                # search for default and enum value properties
                default_property = None
                # # values not accessible for now (25 R1)
                # enum_property = None
                for prop in att.ann_properties:
                    if prop.name == 'NT_DEFAULT_VALUE':
                        default_property = prop
                    # values not accessible for now (25 R1)
                    # elif prop.name == 'NT_ENUM_VALUES':
                    #     enum_property = prop
                # default
                default_value = ''
                ft = FT(att.type)
                if ft == FT.INTEGER:
                    default_value = '0'
                elif ft == FT.REAL:
                    default_value = '0.0'
                elif ft == FT.BOOLEAN:
                    default_value = 'F'
                elif ft == FT.STRING:
                    default_value = ''
                elif ft == FT.TEXT:
                    default_value = ''
                elif ft == FT.CHARACTER:
                    default_value = ' '
                elif ft == FT.DATE:
                    default_value = '01/01/01'
                elif ft == FT.FILE:
                    default_value = ''
                elif ft == FT.ENUM:
                    # values not accessible for now (25 R1)
                    pass

                if default_property:
                    default_value = str(default_property.value)
                    if ft == FT.BOOLEAN:
                        default_value = 'T' if default_value == '1' else 'F'

                self.defaults[att] = default_value

    def create_note(self, annotable: suite.Annotable, note_type: ann.AnnNoteType):
        """Create a default note for an annotable element and a given note type."""
        note = ann.AnnNote(note_type)
        note.ann_note_type = note_type
        note.name = note_type.name
        annotable.ann_notes.append(note)

        for att in note_type.ann_att_definitions:
            ft = FT(att.type)
            if ft == FT.INTEGER:
                cls = ann.AnnAttIntValue
            elif ft == FT.REAL:
                cls = ann.AnnAttRealValue
            elif ft == FT.BOOLEAN:
                cls = ann.AnnAttBoolValue
            elif ft == FT.STRING:
                cls = ann.AnnAttStringValue
            elif ft == FT.TEXT:
                cls = ann.AnnAttStringValue
            elif ft == FT.CHARACTER:
                cls = ann.AnnAttCharValue
            elif ft == FT.DATE:
                cls = ann.AnnAttDateValue
            elif ft == FT.FILE:
                cls = ann.AnnAttFileValue
            elif ft == FT.ENUM:
                cls = ann.AnnAttEnumValue
            else:
                cls = None
            if not cls:
                print(f'unexpected field type {ft.name} for {note_type.name}::{att.name}')
            else:
                default_value = self.defaults[att]
                value = cls(note_type)
                value.ann_att_definition = att
                value.from_string(default_value)
                note.ann_att_values.append(value)

    def visit_annotable(self, annotable: suite.Annotable, *args):
        """Create missing notes of the visited annotable element."""
        # loop on annotabilities
        category = self.get_category(annotable)
        ann_category = self.categories.get(category)
        if ann_category:
            for annotability in ann_category.annotabilities:
                if not annotability.default_note:
                    continue
                note_type = annotability.ann_note_type
                for note in annotable.ann_notes:
                    if note.ann_note_type == note_type:
                        # note already exists
                        break
                else:
                    # note not found: create one with default values
                    print(f'creating {note_type.name} note for {annotable.get_full_path()}')
                    self.create_note(annotable, note_type)
                    st = annotable.defined_in
                    # root tree diagrams do not have an associated file
                    if st:
                        # the ann file has to be updated
                        st.ann_modified = True
                        # the scade file needs to be saved if the notes are embedded as comments
                        st.sao_modified = True

        # not need to call base function
        # super().visit_annotable(annotable, *args)

    def main(self, session: suite.Session) -> int:
        """Entry point for unit testing or reuse."""
        model = session.model
        self.build_categories(model)
        self.compute_default_values(model)
        # visit all model annotable instances and create missing notes
        self.visit(model)
        # save all modified model and notes files (XSCADE, ANN)
        session.save_model2()
        return 0


def main() -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    return_code = 0
    for session in get_sessions():
        code = DefaultNotes().main(session)
        if code != 0:
            return_code = code

    return return_code



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Change duplicate oids."""

import argparse

from ansys.scade.apitools import declare_project

# isort: split
# must be imported after ansys.scade.apitools
from ansys.scade.ps.change_oids.find_duplicates import main as find_duplicates_main
from ansys.scade.ps.change_oids.new_oids import main as new_oids_main

tool = 'Ansys SCADE Power Scripts: Change duplicate OIDs'


def main():
    """Implement ``ansys.scade.ps.change_oids.__main__:main`` packages's script."""
    parser = argparse.ArgumentParser(description=tool)
    subparsers = parser.add_subparsers(help='change_oids sub-commands')

    # find
    find_parser = subparsers.add_parser('find', help='Find duplicate OIDs')
    find_parser.add_argument(
        '-p',
        '--projects',
        metavar='<project>',
        help='SCADE Suite projects',
        required=True,
        nargs='+',
    )
    find_parser.add_argument(
        '-x',
        '--extension',
        metavar='<extension>',
        help='result file extension',
        default='.dupoids.txt',
    )
    find_parser.add_argument(
        '-d', '--dump_paths', help='print model element paths', action='store_true'
    )
    find_parser.set_defaults(cmd='find')

    # change
    change_parser = subparsers.add_parser('new', help='Create new OIDs')
    change_parser.add_argument(
        '-p', '--project', metavar='<project>', help='SCADE Suite project', required=True
    )
    change_parser.add_argument(
        '-f', '--file', metavar='<input file>', help='input file', required=True
    )
    change_parser.add_argument(
        '-m', '--map', metavar='<output map file>', help='output map file', required=True
    )
    change_parser.add_argument('-l', '--log', metavar='<log file>', help='log file', default='')
    change_parser.set_defaults(cmd='change')

    options = parser.parse_args()

    assert declare_project  # nosec B101  # declare_project must be defined on Windows
    # cmd is either find or change
    if options.cmd == 'find':
        for project in options.projects:
            declare_project(project)
        code = find_duplicates_main(options.extension, options.dump_paths)
    else:
        # assert options.cmd == 'change'
        declare_project(options.project)
        code = new_oids_main(options.file, options.map, options.log)
    exit(code)


if __name__ == '__main__':
    # run with python.exe -m ansys.scade.ps.change_oids
    main()
# else:  # run with ansys_scade_ps_change_oids.exe



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Rename unused files."""

from pathlib import Path

from scade.model.project.stdproject import Project, get_roots as get_projects

tool = 'Ansys SCADE Power Scripts: Rename unused files'


class RenameUnused:
    """Renames unused files (SCADE, XSCADE, ANN)."""

    def main(self, project: Project) -> int:
        """Entry point for unit testing or reuse."""
        # cache files
        file_refs = [Path(_.pathname) for _ in project.file_refs]
        scade_files = {_ for _ in file_refs if _.suffix.lower() in ['.scade', '.xscade']}
        dir = Path(project.pathname).parent.absolute()
        # model files
        for pattern in ['*.scade', '*.xscade']:
            for file in dir.glob(pattern):
                if file not in scade_files:
                    self.rename(file)
        # annotation files
        for file in dir.glob('*.ann'):
            # check xscade files only: scade files do not have annotations
            model_file = file.with_suffix('.xscade')
            if model_file not in scade_files:
                self.rename(file)
        return 0

    def rename(self, file: Path):
        """Suffix the file name with ``.toremove``."""
        print(f'renaming {file.name}')
        file.rename(file.with_name(file.name + '.toremove'))


def main() -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    return_code = 0
    for project in get_projects():
        code = RenameUnused().main(project)
        if code != 0:
            return_code = code

    return return_code



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Ansys SCADE Power Scripts: Set equation sets styles.

Reset all the styles of diagrams' equation sets

.. Note::

  * The script can not be executed on the command line.
  * Each equation set is assigned the style ``<prefix><count>`` where
    ``<count>`` starts from 1 and is reset for each diagram, <prefix> is the
    ``PREFIX`` constant defined at the beginning of the script.
  * The <prefix><number> styles are not created.
  * The equation sets are given a default style if the requested one does
    not exist.
"""

import scade.model.suite as suite
from scade.model.suite import get_roots as get_sessions
from scade.model.suite.visitors import Visit

# prefix for styles
PREFIX = 'EQS'


class SetStyle(Visit):
    """Visitor for equation sets."""

    def __init__(self):
        """Visit all loaded models."""
        for session in get_sessions():
            model = session.model
            self.visit(model)

    def visit_net_diagram(self, net_diagram: suite.NetDiagram, *args):
        """Reset the counter before a graphical diagram is visited."""
        self.count = 0
        super().visit_net_diagram(net_diagram, *args)

    def visit_equation_set(self, equation_set: suite.EquationSet, *args):
        """Create or update the style of the visited equation set."""
        self.count += 1
        # suite.set_style is defined dynamically when executed from the SCADE IDE
        suite.set_style(equation_set, '%s%d' % (PREFIX, self.count))  # type: ignore  # reportAttributeAccessIgnore


SetStyle()



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Copy SCADE project properties."""

import json
import os
from pathlib import Path
from typing import Dict, List

from scade.model.project.stdproject import Project, get_roots as get_projects

import ansys.scade.apitools.create as create


class CopyProperties:
    """Copy tool properties from a reference projects to other ones."""

    def __init__(self, schema: str):
        # options
        self.schema = Path(schema)

    def main(self, reference: Project, projects: List[Project]) -> int:
        """Entry point for unit testing or reuse."""
        # read the schema
        try:
            schema = json.loads(self.schema.read_text())
        except BaseException as e:
            print(str(e))
            return 1

        # create a dictionaries of properties with the 'relative' attribute
        tool_props = {}
        for tool, props in schema.items():
            for prop in props:
                if prop[0] == '@':
                    prop = prop[1:]
                    relative = True
                else:
                    relative = False
                tool_prop = f'@{tool}:{prop}'
                tool_props[tool_prop] = relative

        # propagate the selected properties from the reference to the other projects
        modified_projects = []
        for project in projects:
            if self.copy_properties(reference, project, tool_props):
                modified_projects.append(project)

        # save modified projects, once everything is done
        for project in modified_projects:
            create.save_project(project)

        return 0

    def copy_properties(self, src: Project, dst: Project, tool_props: Dict[str, bool]) -> bool:
        """Copy the selected properties from src to dst and return whether dst is modified."""
        # return value
        modified = False
        # local caches in the destination project: key = prop.name, configuration.name
        props = {}
        for dst_prop in dst.props:
            if dst_prop.name in tool_props:
                configuration = dst_prop.configuration.name if dst_prop.configuration else ''
                # index the property by its name and configuration
                props[dst_prop.name, configuration] = dst_prop
        configurations = {_.name: _ for _ in dst.configurations}

        # propagate the properties
        for src_prop in src.props:
            prop_name = src_prop.name
            if prop_name not in tool_props:
                continue
            relative = tool_props[prop_name]

            # check the existence of the configuration in the destination project
            if src_prop.configuration:
                conf_name = src_prop.configuration.name
                dst_conf = configurations.get(conf_name)
                if not dst_conf:
                    dst_conf = create.create_configuration(dst, conf_name)
                    configurations[conf_name] = dst_conf
                    modified = True
            else:
                dst_conf = None
                conf_name = ''

            # update the property's values
            values = src_prop.values
            if relative:
                src_dir = Path(src.pathname).parent
                dst_dir = Path(dst.pathname).parent
                new_values = []
                for pathname in values:
                    if pathname and pathname[0] != '$':
                        path = Path(pathname)
                        # make the path absolute wrt the source project
                        path = (src_dir / path).resolve()
                        # make the path relative wrt the destination project
                        path = os.path.relpath(path, start=dst_dir)
                        pathname = str(path)
                    new_values.append(pathname)

                values = new_values

            key = prop_name, conf_name
            dst_prop = props.get(key)
            if not dst_prop:
                # create the property
                print(f'creating {prop_name} for {conf_name}')
                dst_prop = create.create_prop(dst, dst_conf, prop_name, values)
                modified = True
            else:
                if dst_prop.values != values:
                    print(f'updating {prop_name} for {conf_name} with {values}')
                    dst_prop.values = values
                    modified = True
                # remove the property from the cache, remaining ones should be deleted
                props.pop(key)

        # delete unused properties, for example properties having default values
        for dst_prop in props.values():
            dst.props.remove(dst_prop)
            modified = True

        return modified


def main(reference: str, schema: str) -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    ref_project = None
    projects = []
    for project in get_projects():
        if Path(project.pathname).name == reference:
            ref_project = project
        else:
            projects.append(project)
    if not ref_project:  # pragma: nocover
        # this may happen if and only if the script is run with ``scade.exe -script``
        print(f'{reference}: project not found')
        return 1
    code = CopyProperties(schema).main(ref_project, projects)
    return code



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Obfuscator for Scade models."""

import argparse

from ansys.scade.apitools import declare_project

# isort: split
# must be imported after ansys.scade.apitools
from ansys.scade.ps.obfuscator.obfuscator import main as obfuscator_main, tool


def main():
    """Implement ``ansys.scade.ps.obfuscator.__main__:main`` packages's script."""
    parser = argparse.ArgumentParser(description=tool)
    parser.add_argument(
        '-p', '--project', metavar='<project>', help='SCADE Suite project', required=True
    )
    parser.add_argument(
        '-t', '--trace', metavar='<trace>', help='Output trace file', required=False
    )
    parser.add_argument('-i', '--internals', action='store_true', help='Rename internal variables')
    parser.add_argument(
        '-l', '--ignored_libraries', metavar='library', nargs='*', help='Ignored libraries'
    )
    parser.add_argument(
        '-s', '--seed', metavar='<seed>', help='Random seed for obfuscation', required=False
    )
    options = parser.parse_args()

    assert declare_project  # nosec B101  # declare_project must be defined on Windows
    declare_project(options.project)
    code = obfuscator_main(
        options.trace, options.internals, options.ignored_libraries, options.seed
    )
    exit(code)


if __name__ == '__main__':
    # run with python.exe -m ansys.scade.ps.obfuscator
    main()



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Obfuscator for Scade models."""

from pathlib import Path

# nosec B311  # usage of random is acceptable for obfuscation
# it allows to provide seeds for reproducible unit tests.
import random
from typing import Dict, List, Optional

import scade
import scade.model.suite as suite
from scade.model.suite import get_roots as get_sessions
from scade.model.suite.visitors import Visit

tool = 'Ansys SCADE Power Scripts: Obfuscator for Scade models'


def _get_scade_path(object_: suite.Object) -> str:
    """Return a path for objects for which get_full_path is not enough."""
    if isinstance(object_, suite.CompositeElement):
        # structure's field
        return object_.owner.get_full_path() + object_.name
    elif isinstance(object_, suite.Diagram):
        return object_.owner.get_full_path() + object_.name
    elif isinstance(object_, suite.EquationSet):
        return _get_scade_path(object_.owner) + '/' + object_.name
    else:
        return object_.get_full_path()


class _State:
    """Global variables for the obfuscation process."""

    def __init__(self):
        # dictionary old name to new name
        self.names = {}  # type: Dict[str, str]
        # dictionary object to old path
        self.old_names = {}  # type: Dict[suite.Object, str]
        # dictionary object to new path
        self.new_names = {}  # type: Dict[suite.Object, str]
        # modified storage units
        self.units = []  # type: List[suite.StorageUnit]
        # dictionary old file name to new file name
        self.files = {}  # type: Dict[str, str]


class _FilteredVisit(Visit):
    """Base visitor for caching and renaming items."""

    def __init__(self, state: _State, rename_internals: bool, ignored_libraries: List[str]):
        self.state = state
        self.libraries = set(ignored_libraries)
        self.rename_internals = rename_internals

    def visit(self, item):
        """Redefinition to consider instances of Session."""
        if isinstance(item, suite.Session):
            # visit all loaded models
            for model in item.loaded_models:
                super().visit(model)
        else:
            super().visit(item)

    def visit_object(self, object_: suite.Object, *args):
        """Call ``on_named`` if the visited object has a name."""
        if hasattr(object_, 'name') and object_.name:
            # process the named object
            self.on_named(object_)

    def visit_local_variable(self, local_variable: suite.LocalVariable, *args):
        """Filter internal variables depending on the options."""
        if not local_variable.is_internal() or self.rename_internals:
            super().visit_local_variable(local_variable, *args)
        # else: skip internal variables

    def on_named(self, named: suite.Object):
        """Process a named object."""
        pass


class _Cache(_FilteredVisit):
    """Visitor to cache the objects' path before they are renamed."""

    def on_named(self, named: suite.Object):
        """Process a named object."""
        if isinstance(named, suite.Model):
            self.state.old_names[named] = named.name
        else:
            self.state.old_names[named] = _get_scade_path(named)

    def visit_model(self, model: suite.Model, *args):
        """Filter the libraries that should be ignored."""
        if model.name not in self.libraries:
            super().visit_model(model, *args)


class _Rename(_FilteredVisit):
    """Visitor to rename objects and files."""

    def get_name(self, old_name: str) -> str:
        """
        Create a new name: pattern XXX999, random characters and digits.

        The mapping is stored in a dictionary, to make sure there
        is only one new name for an existing one.
        """
        name = self.state.names.get(old_name, '')
        while not name:
            name = ''.join(
                ['%c' % (65 + random.randrange(0, 26)) for _ in range(3)]  # nosec B311
                + ['%c' % (48 + random.randrange(0, 10)) for _ in range(3)]  # nosec B311
            )
            # check uniqueness
            if name in self.state.names:
                # unlikely but who knows?
                # same player shoot again
                name = ''
            else:
                if old_name and old_name[0] == "'":
                    # polymorphic type
                    name = "'" + name
                self.state.names[old_name] = name

        return name

    def on_named(self, named: suite.Object):
        """Process a named object."""
        named.name = self.get_name(named.name)
        if isinstance(named, suite.Model):
            self.state.new_names[named] = named.name
        else:
            # name of owning objects already changed
            self.state.new_names[named] = _get_scade_path(named)

    def visit_const_value(self, const_value: suite.ConstValue, *args):
        """Rename the expression's label if any."""
        if const_value.kind == 'Label':
            # label for projections
            const_value.value = self.get_name(const_value.value)
        super().visit_const_value(const_value, *args)

    def visit_storage_unit(self, storage_unit: suite.StorageUnit, *args):
        """Delete the referenced files and mark them as modified."""

        def with_stem(path: Path, name: str) -> Path:
            """Implement Path.with_stem for Python 3.7."""
            return path.with_name(name + path.suffix)

        path_sao = Path(storage_unit.sao_file_name)
        new_name = self.get_name(path_sao.stem)
        if path_sao.exists():
            path_sao.unlink()
            storage_unit.sao_modified = True
        path_ann = Path(storage_unit.ann_file_name)
        if path_ann.exists():
            # for some reason, it is not possible to save the annotation files
            # - StorageUnit.save does not save annotation files
            # - Session.save_model2 does not save libraries
            # so rename the annotation files since they are not modified
            path_ann.rename(with_stem(path_ann, new_name))
        self.state.units.append(storage_unit)
        # path of the file
        storage_unit.sao_file_name = str(with_stem(path_sao, new_name))
        storage_unit.ann_file_name = str(with_stem(path_ann, new_name))
        # reference of the file
        persist_as = Path(storage_unit.persist_as)
        new_persist_as = with_stem(persist_as, new_name)
        storage_unit.persist_as = str(new_persist_as)
        # record the name change
        suffix = path_sao.suffix
        self.state.files[persist_as.with_suffix(suffix).name] = new_persist_as.with_suffix(
            suffix
        ).name
        super().visit_storage_unit(storage_unit, *args)

    def visit_presentation_element(self, presentation_element: suite.PresentationElement, *args):
        """Remove formatting, if any."""
        presentation_element.text_areas = []
        super().visit_presentation_element(presentation_element, *args)

    def visit_text_diagram(self, text_diagram: suite.TextDiagram, *args):
        """Remove formatting, if any."""
        # None is a legal value to remove an association
        text_diagram.text_area = None  # type: ignore
        super().visit_text_diagram(text_diagram, *args)

    def visit_model(self, model: suite.Model, *args):
        """Rename the items of a model if it is not excluded."""
        if model.name not in self.libraries:
            # initialize an entry for the model and
            # register the name change for the project file
            new_name = self.get_name(model.name)
            self.state.files[model.name + '.etp'] = new_name + '.etp'
            super().visit_model(model, *args)
        else:
            # register the storage units for saving only
            self.state.units += model.model_storage_units


class Obfuscator:
    """Obfuscation of a model."""

    def __init__(
        self, trace_file: str, rename_internals: bool, ignored_libraries: List[str], seed: str
    ):
        # options
        self.trace_file = trace_file
        self.rename_internals = rename_internals
        self.ignored_libraries = ignored_libraries
        if seed:
            random.seed(seed)
        # runtime caches
        self.state = _State()

    def update_projects(self, session: suite.Session):
        """Propagate the filename changes for each project."""
        # two passes to avoid computing the project's dependencies
        projects = []
        # first pass to update the file references
        for model in session.loaded_models:
            path = Path(model.descriptor.model_file_name)
            # scade is a CPython module defined dynamically
            # scade is a CPython module defined dynamically
            project = scade.load_project(str(path))  # type: ignore
            # referenced files
            for file_ref in project.file_refs:
                path_file = Path(file_ref.pathname)
                new_name = self.state.files.get(path_file.name)
                if new_name:
                    file_ref.set_path_name(str(path_file.with_name(new_name)))
            # project
            if path.stem in self.ignored_libraries:
                # project not renamed but might reference renamed files
                project.save(str(path))
            else:
                new_name = self.state.files.get(path.name, path.name)
                projects.append((project, path.with_name(new_name)))
        # second pass to save the renamed projects
        for project, path in projects:
            path_project = Path(project.pathname)
            if path_project.exists():
                # missing_ok not available Python 3.7
                path_project.unlink()
            project.save(str(path))

    def dump_traceability(self, f, src: Dict[suite.Object, str], dst: Dict[suite.Object, str]):
        """Report the tracebility between old names and new names."""
        keys = sorted([(key, value) for key, value in src.items()], key=lambda _: _[1])
        for key, name in keys:
            f.write('%s\t%s\n' % (name, dst[key]))

    def main(self, session: suite.Session) -> int:
        """Entry point for unit testing or reuse."""
        # cache all objects with old names
        self.cache_items(session)
        # rename all objects
        self.rename_items(session)
        # save the units
        for unit in self.state.units:
            unit.save()
        # update the projects
        self.update_projects(session)
        # dump mapping
        if self.trace_file:
            with Path(self.trace_file).open('w') as f:
                f.write('- old names to new names\n')
                self.dump_traceability(f, self.state.old_names, self.state.new_names)
                f.write('- new names to old names\n')
                self.dump_traceability(f, self.state.new_names, self.state.old_names)

        print('command completed')
        return 0

    def cache_items(self, session):
        """Return the dictionary object to scade path before the renamings."""
        visitor = _Cache(self.state, self.rename_internals, self.ignored_libraries)
        visitor.visit(session)

    def rename_items(self, session):
        """Rename the loaded models and return the dictionary object to scade path."""
        visitor = _Rename(self.state, self.rename_internals, self.ignored_libraries)
        visitor.visit(session)


def main(
    trace_file: str = '',
    rename_internals: bool = False,
    ignored_libraries: Optional[List[str]] = None,
    seed: str = '',
) -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    session = get_sessions()[0]
    if ignored_libraries is None:
        ignored_libraries = []
    code = Obfuscator(trace_file, rename_internals, ignored_libraries, seed).main(session)
    return code



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Update traceability with respect to new OIDs."""

from pathlib import Path
from typing import Dict, List

from lxml import etree as et


class LLR:
    """Entry of a trace file (ALMGT)."""

    def __init__(self, id: str = '', path: str = ''):
        self.elem = None
        self.id = id
        self.path = path
        self.adds = {}
        self.rems = {}

    def is_empty(self):
        """Return whether the LLR has links to add or remove."""
        return len(self.adds) + len(self.rems) == 0

    def parse(self, elem):
        """Parse the XML element's hierarchy."""
        self.elem = elem
        self.id = self.elem.get('id')
        self.path = self.elem.get('pathName', '')
        for elem in self.elem.findall('requirement'):
            req = elem.get('id')
            kind = elem.get('traceType')
            if kind == 'ADD_LINK':
                self.adds[req] = elem
            else:
                self.rems[req] = elem
        return self

    def update_tree(self):
        """Create the missing xml elements."""
        for req, elem in self.adds.items():
            if elem is None:
                et.SubElement(self.elem, 'requirement', {'id': req, 'traceType': 'ADD_LINK'}, None)

        for req, elem in self.rems.items():
            if elem is None:
                et.SubElement(
                    self.elem, 'requirement', {'id': req, 'traceType': 'REMOVE_LINK'}, None
                )

    def remove(self, req: str) -> bool:
        """Remove a link to a requirement."""
        if req in self.rems:
            # requirement being removed: nothing to do
            return False
        elif req in self.adds:
            # new requirement: remove it from the list
            elem = self.adds.pop(req)
            assert self.elem is not None  # nosec B101  # addresses linter
            self.elem.remove(elem)
        else:
            # register for deletion
            self.rems[req] = None
        return True

    def add(self, req: str):
        """Add a link to a requirement."""
        self.adds[req] = None


class GTFile:
    """Content of SCADE ALM Gateway traceability cache file(ALMGT)."""

    def __init__(self):
        self.tree = None
        self.llrs = {}

    def parse(self, trace: Path) -> bool:
        """Parse the trace file."""
        parser = et.XMLParser(remove_blank_text=True)
        try:
            self.tree = et.parse(str(trace), parser)
        except OSError as e:
            print(e)
            return False
        for elem in self.tree.getroot().findall('object'):
            llr = LLR().parse(elem)
            self.llrs[llr.id] = llr
        return True

    def save(self, trace: Path):
        """Save the trace file."""
        assert self.tree is not None  # nosec B101  # addresses linter
        root = self.tree.getroot()
        for llr in self.llrs.values():
            if llr.is_empty():
                if llr.elem is not None:
                    root.remove(llr.elem)
                continue
            if llr.elem is None:
                # sounds like path is no more present in traceability files (ALMGT)
                # llr.elem = et.SubElement(root, 'object', {'id': llr.id, 'path': llr.path}, None)
                llr.elem = et.SubElement(root, 'object', {'id': llr.id}, None)
            llr.update_tree()
        self.tree.write(
            str(trace), encoding='utf-8', standalone='yes', xml_declaration=True, pretty_print=True
        )

    def replace_oid(self, subst: Dict[str, str]):
        """Add an entry to replace an oid if the element traces requirements."""
        reqs = subst['reqs']
        if not reqs:
            return
        path = subst['path']
        old_oid = subst['old_oid']
        new_oid = subst['new_oid']
        old_llr = self.llrs.setdefault(old_oid, LLR(old_oid, path))
        new_llr = LLR(new_oid, path)
        self.llrs[new_oid] = new_llr
        for req in reqs:
            if old_llr.remove(req):
                new_llr.add(req)


def update_trace(substs: List[Dict[str, str]], trace: Path) -> bool:
    """Update trace file (ALMGT) with the changes recorded in map_oids."""
    if not trace.exists():
        # load the almgt template file, same directory as the script
        source = Path(__file__).parent / 'template.almgt'
    else:
        source = trace
    gt = GTFile()
    if gt.parse(source):
        for subst in substs:
            gt.replace_oid(subst)
        gt.save(trace)
        return True
    return False



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Change duplicate OIDs."""

from pathlib import Path
from typing import List, Set

import scade.model.suite as suite
from scade.model.suite import get_roots as get_sessions
from scade.model.suite.visitors import Visit


class CacheOids(Visit):
    """Visitor to cache the duplicated OIDs."""

    def __init__(self, oids: Set[str], model: suite.Model):
        # global set of oids
        self.oids = oids
        self.model = model
        # duplicated oids
        self.duplicates = []  # type: List[suite.Annotable]
        # run the visit, model only: no libraries
        self.visit(model)

    def visit_annotable(self, annotable: suite.Annotable, *args):
        """Cache the object's oid and check whether it already exists."""
        oid = annotable.get_oid()
        if oid:
            # ignore instances without oid, like array or structures
            if oid in self.oids:
                self.duplicates.append(annotable)
            else:
                self.oids.add(oid)
        super().visit_annotable(annotable, *args)


class FindDuplicates:
    """Find duplicate OIDs for a set of SCADE Suite models."""

    def __init__(self, extension: str, dump_paths: bool):
        # options
        self.extension = extension
        self.dump_paths = dump_paths

    def main(self, sessions: List[suite.Session]) -> int:
        """Entry point for unit testing or reuse."""
        code = 0
        # global set of oids
        oids = set()
        # find duplicate oids
        caches = [CacheOids(oids, session.model) for session in sessions]
        # duymp the results
        for cache in caches:
            path = Path(cache.model.project.pathname).with_suffix(self.extension)
            try:
                f = path.open('w')
            except Exception as e:
                print(str(e))
                # an error occurred
                code = 1
                continue
            # actually, oid is useless
            for item in cache.duplicates:
                tokens = [item.get_oid()]
                if self.dump_paths:
                    tokens.append(self.get_path(item))
                f.write('%s\n' % '\t'.join(tokens))
            f.close()

        return code

    def get_path(self, object_: suite.Object) -> str:
        """
        Provide a Scade path for objects that don't have one.

        For objects that don't have a path, the function returns their name
        appended to their owner's path. Otherwise, returns their own path.
        """
        if isinstance(object_, suite.TreeDiagram):
            path = object_.owner.get_full_path() + f'<{object_.block_kind}>'
        elif isinstance(object_, suite.CompositeElement):
            return object_.owner.get_full_path() + f'<{object_.name}>'
        else:
            path = object_.get_full_path()

        return path


def main(extension: str, use_paths: bool) -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    code = FindDuplicates(extension, use_paths).main(get_sessions())
    return code



# -*- coding: utf-8 -*-

# Copyright (C) 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Ansys SCADE Power Scripts: Change duplicate OIDs."""

import json
from pathlib import Path

import scade.model.suite as suite
from scade.model.suite import get_roots as get_sessions
from scade.model.suite.visitors import Visit

try:
    import scade.model.traceability as traceability
except ModuleNotFoundError:
    # module exists only since SCADE 2024 R1
    traceability = None

from ansys.scade.ps.change_oids.update_trace import update_trace


class _Logger:
    """Logging support."""

    def __init__(self, file: str):
        self.f = Path(file).open('w') if file else None

    def log(self, *args):
        """Log a message either to a file or the standard output."""
        if self.f:
            self.f.write(' '.join(args) + '\n')
        else:
            print(*args)


class NewOids(Visit):
    """Ansys SCADE Power Scripts: Change duplicate OIDs."""

    def __init__(self, file: str, map: str, log: str):
        self.file = Path(file)
        self.map = Path(map)
        self.log = _Logger(log)
        # cache
        self.map_oids = {}  # Dict[str, suite.Annotable]
        # substitutions done
        self.substs = []  # List[Dict[str, str]]
        # requirements
        self.map_traceability_links = {}  # Dict[str, List[str]]

    def main(self, session: suite.Session) -> int:
        """Entry point for unit testing or reuse."""
        # index the model elements by oid
        self.visit(session.model)
        # read the duplicated oids: ignore comments and optional paths
        oids = [_.split('\t')[0] for _ in self.file.read_text().split('\n') if _ and _[0] != '#']

        # read the requirements, if any
        path = Path(session.model.project.pathname)
        if traceability:
            try:
                req_project = traceability.load(str(path))
            except BaseException:
                req_project = None
        else:
            req_project = None
        if req_project:
            for element in req_project.traceable_elements:
                self.map_traceability_links[element.identifier] = [
                    _.target.identifier for _ in element.outgoing_links
                ]
        else:
            self.map_traceability_links = {}
        # compute new oids
        for oid in oids:
            annotable = self.map_oids.get(oid)
            if not annotable:
                self.log.log(f'{oid}: model element not found')
            else:
                if not self.process_element(annotable):
                    # no need to continue
                    return 1

        # update the traceability links (ALMGT)
        trace = Path(session.model.project.pathname).with_suffix('.almgt')
        if not update_trace(self.substs, trace):
            return 1

        # dump the mapping
        try:
            with self.map.open('w') as f:
                json.dump(self.substs, f, indent=4, sort_keys=True)
        except BaseException as e:
            self.log.log(str(e))
            return 1

        # save modified model and annotation files
        session.save_model2()

        return 0

    def process_element(self, annotable: suite.Annotable) -> bool:
        """Change the oid of a model element and record updates."""
        self.log.log(f'updating oid for {self.get_path(annotable)}')
        old_oid = annotable.get_oid()
        # create a new oid using a ghost object, for example a constant
        new_oid = suite.Constant().get_oid()
        # requirement_ids not available for scripts executed outside of the SCADE IDE
        # reqs = annotable.requirement_ids if isinstance(annotable, suite.Traceable) else []
        reqs = self.map_traceability_links.get(annotable.get_oid(), [])
        self.substs.append(
            {
                'path': self.get_path(annotable),
                'old_oid': old_oid,
                'new_oid': new_oid,
                'reqs': reqs,
            }
        )

        try:
            # available only in recent releases of SCADE Suite (starting 2021 R2?)
            annotable.set_oid(new_oid)
        except BaseException as e:
            self.log.log(str(e))
            return False
        unit = annotable.defined_in
        unit.sao_modified = True
        if isinstance(annotable, suite.Annotable) and annotable.ann_notes:
            unit.ann_modified = True

        return True

    def get_path(self, object_: suite.Object) -> str:
        """
        Provide a Scade path for objects that don't have one.

        For objects that don't have a path, the function returns their name
        appended to their owner's path. Otherwise, returns their own path.
        """
        if isinstance(object_, suite.TreeDiagram):
            path = object_.owner.get_full_path() + f'<{object_.block_kind}>'
        elif isinstance(object_, suite.CompositeElement):
            return object_.owner.get_full_path() + f'<{object_.name}>'
        else:
            path = object_.get_full_path()

        return path

    def visit_annotable(self, annotable: suite.Annotable, *args):
        """Index the model element by oid."""
        self.map_oids[annotable.get_oid()] = annotable
        super().visit_annotable(annotable, *args)


def main(file: str, map: str, log: str = '') -> int:
    """Entry point for ``scade.exe -script`` or called by ``__main__``."""
    code = NewOids(file, map, log).main(get_sessions()[0])
    return code
